(* Corrig des exercices de programmation du chapitre 3 *)

(* Le code donn ci-dessous est trs (trop) fidle  la description donne
   dans les notes de cours: il utilise une reprsentation inefficace des
   substitutions, effectue l'occur-check de faon coteuse, calcule l'ensemble
   des variables libres dans l'environnement de faon nave, etc. On peut
   implmenter l'infrence de types pour ML de faon beaucoup plus efficace,
   mais la correspondance avec les rgles de typage est alors plus difficile
    justifier. Ce pourrait tre le sujet d'un cours plus avanc. *)

(********** Exercice 3.1: propagation des types pour ML monomorphe *********)

(* L'algbre de types (comme au chapitre 1) *)

type exprtype =
    Type_base of string
  | Type_var of string
  | Type_flche of exprtype * exprtype
  | Type_produit of exprtype * exprtype

(* Type des oprateurs (monomorphes) *)

let type_oprateur op =
  match op with
    "+" | "-" | "*" | "/" ->
      Type_flche(Type_produit(Type_base "int", Type_base "int"),
                  Type_base "int")
  | "fix" ->
      let int_f_int = Type_flche(Type_base "int", Type_base "int") in
      Type_flche(Type_flche(int_f_int, int_f_int), int_f_int)
  | _ ->
      assert false (* oprateur non support *)

module Exercice31 = struct

(* La syntaxe abstraite des expressions mini-ML avec annotation de type
   sur les paramtres des "fun" *)

type expression =
    Var of string
  | Const of int
  | Op of string
  | Fun of string * exprtype * expression
  | App of expression * expression
  | Paire of expression * expression
  | Fst of expression
  | Snd of expression
  | Let of string * expression * expression

(* La fonction de propagation de types.

   Noter que dans le code ci-dessous, si on trouve une variable alors qu'on attend
   une flche ou un produit, on signale une erreur. On ne fait donc aucune forme
   d'infrence. En fait, les variables de types sont traites ici exactement comme
   des types de base inconnus. *)

exception Erreur_de_type

let rec type_de env a =
  match a with
    Var x -> List.assoc x env
  | Const c -> Type_base "int"
  | Op op -> type_oprateur op
  | Fun(x, tau, b) -> Type_flche(tau, type_de ((x, tau) :: env) b)
  | App(b, c) ->
      begin match type_de env b with
        Type_flche(ty_arg, ty_rs) ->
          if ty_arg = type_de env c then ty_rs else raise Erreur_de_type
      | _ -> raise Erreur_de_type
      end
  | Paire(b, c) -> Type_produit(type_de env b, type_de env c)
  | Fst b ->
      begin match type_de env b with
      |	Type_produit (t1, _) -> t1
      |	_ -> raise Erreur_de_type
      end
  | Snd b ->
      begin match type_de env b with
      |	Type_produit (_, t2) -> t2
      |	_ -> raise Erreur_de_type
      end
  | Let(x, b, c) ->
      let ty_b = type_de env b in
      type_de ((x, ty_b) :: env) c

end (* Exercice31 *)

(****************** Exercice 3.2: construction de C(a) *********************)

(* La syntaxe abstraite des expressions mini-ML (comme au chapitre 1) *)

type expression =
    Var of string
  | Const of int
  | Op of string
  | Fun of string * expression
  | App of expression * expression
  | Paire of expression * expression
  | Fst of expression
  | Snd of expression
  | Let of string * expression * expression

(* Gnration de nouvelles variables de types a1, a2, a3, ... *)

let compteur_de_variables = ref 0

let nouvelle_variable _ =
  incr compteur_de_variables;
  "a" ^ string_of_int !compteur_de_variables

let nouveau _ =
  Type_var (nouvelle_variable ())

(* Associer de nouvelles variables de type  chaque identificateur et
    chaque sous-expression d'une expression *)

type identificateur_typ =
  { id : string; type_id: exprtype }

type expression_type = 
  { descr: expression_type_descr; type_expr: exprtype }

and expression_type_descr =
    TVar of identificateur_typ
  | TConst of int
  | TOp of string
  | TFun of identificateur_typ * expression_type
  | TApp of expression_type * expression_type
  | TPaire of expression_type * expression_type
  | TFst of expression_type
  | TSnd of expression_type
  | TLet of identificateur_typ * expression_type * expression_type

let rec annotation env expr =
  let descr =
    match expr with
      Var x -> TVar(List.assoc x env)
    | Const c -> TConst c
    | Op op -> TOp op
    | Fun(x, a) ->
        let tx = {id = x; type_id = nouveau()} in
        TFun(tx, annotation ((x, tx) :: env) a)
    | App(a, b) ->
        TApp(annotation env a, annotation env b)
    | Paire(a, b) ->
        TPaire(annotation env a, annotation env b)
    | Fst a ->
	TFst (annotation env a)
    | Snd a ->
	TSnd (annotation env a)
    | Let(x, a, b) ->
        let tx = {id = x; type_id = nouveau()} in
        TLet(tx, annotation env a, annotation ((x, tx) :: env) b)
  in
    { descr = descr; type_expr = nouveau() }

(* Gnration de l'ensemble d'quations C(a) pour une expression annote a *)

let rec quations a =
  match a.descr with
    TVar id -> [a.type_expr, id.type_id]
  | TConst c -> [a.type_expr, Type_base "int"]
  | TOp op -> [a.type_expr, type_oprateur op]
  | TFun(id, b) ->
      (a.type_expr, Type_flche(id.type_id, b.type_expr)) :: quations b
  | TApp(b, c) ->
      (b.type_expr, Type_flche(c.type_expr, a.type_expr)) ::
      quations b @ quations c
  | TPaire(b, c) ->
      (a.type_expr, Type_produit(b.type_expr, c.type_expr)) ::
      quations b @ quations c
  | TFst b ->
      (b.type_expr, Type_produit(a.type_expr, nouveau())) ::
      quations b
  | TSnd b ->
      (b.type_expr, Type_produit(nouveau(), a.type_expr)) ::
      quations b
  | TLet(x, b, c) ->
      (a.type_expr, c.type_expr) ::
      (x.type_id, b.type_expr) ::
      quations b @ quations c

(******************** Exercice 3.3: unification **************************)

(* Reprsentation des substitutions par des fonctions des types dans les
   types *)

(* La substitution lmentaire [alpha_i <- tau_i] *)

let subst alpha_tau_list =
  let rec do_subst tau =
    match tau with
      Type_var alpha ->
        begin try
          List.assoc alpha alpha_tau_list
        with Not_found ->
          tau
        end
    | Type_base b -> tau
    | Type_flche(tau1, tau2) -> Type_flche(do_subst tau1, do_subst tau2)
    | Type_produit(tau1, tau2) -> Type_produit(do_subst tau1, do_subst tau2)
  in
    do_subst

(* La composition de substitutions et la substitution identit *)

let compose subst1 subst2 = fun typ -> subst1(subst2(typ))

let identit = fun typ -> typ

(* Appliquer une substitution  un ensemble d'quations entre types
   (une liste de paires de types) *)

let subst_equations sigma c =
  List.map (fun (tau1, tau2) -> (sigma tau1, sigma tau2)) c

(* Test d'occurrence d'une variable dans un type *)

let rec occurrence alpha tau =
  match tau with
    Type_base b -> false
  | Type_var beta -> alpha = beta
  | Type_flche(tau1, tau2) -> occurrence alpha tau1 || occurrence alpha tau2
  | Type_produit(tau1, tau2) -> occurrence alpha tau1 || occurrence alpha tau2

(* Le calcul de l'unificateur principal *)

exception Non_unifiable

let rec mgu = function
    [] -> identit
  | (Type_var alpha, Type_var beta) :: c when alpha = beta -> mgu c
  | (Type_var alpha, tau) :: c when not(occurrence alpha tau) ->
      let s = subst [alpha, tau] in 
      compose (mgu(subst_equations s c)) s
  | (tau, Type_var alpha) :: c when not(occurrence alpha tau) ->
      let s = subst [alpha, tau] in 
      compose (mgu(subst_equations s c)) s
  | (Type_base b1, Type_base b2) :: c when b1 = b2 -> mgu c
  | (Type_flche(tau1, tau2), Type_flche(tau1', tau2')) :: c ->
      mgu((tau1, tau1') :: (tau2, tau2') :: c)
  | (Type_produit(tau1, tau2), Type_produit(tau1', tau2')) :: c ->
      mgu((tau1, tau1') :: (tau2, tau2') :: c)
  | _ ->
      raise Non_unifiable

(******************* Exercice 3.4: l'algorithme I **************************)

type rsultat = Err | Type of exprtype

let infrence_monomorphe a =
  try
    let a' = annotation [] a in
    let phi = mgu(quations a') in
    Type(phi(a'.type_expr))
  with Non_unifiable ->
    Err

(******************** Exercice 3.5: l'algorithme W ***********************)

(* Les schmas de types (comme au chapitre 1) *)

type schma = { quantif: string list; corps: exprtype }

(* Prendre une instance gnrique d'un schma *)

let instance_gnrique sigma =
  (* On gnre les variables nouvelles beta1... betaN *)
  let vars_inst =
    List.map nouveau sigma.quantif in
  (* On construit la substitution alphai <- betai *)
  let s = subst (List.combine sigma.quantif vars_inst) in
  (* On l'applique au corps du schma *)
  s sigma.corps

(* Calcul des variables libres -- comme au chapitre 1 *)

let rec union l1 l2 =
  match l1 with
    [] -> l2
  | hd :: tl -> union tl (if List.mem hd l2 then l2 else hd :: l2)

let rec variables_libres t =
  match t with
    Type_base b -> []
  | Type_var v -> [v]
  | Type_flche(t1, t2) -> union (variables_libres t1) (variables_libres t2)
  | Type_produit(t1, t2) -> union (variables_libres t1) (variables_libres t2)

let rec diffrence l1 l2 =
  match l1 with
    [] -> []
  | hd :: tl ->
      if List.mem hd l2 then diffrence tl l2 else hd :: diffrence tl l2

let variables_libres_schma s =
  diffrence (variables_libres s.corps) s.quantif

let rec variables_libres_env env =
  match env with
    [] -> []
  | (ident, schma) :: reste ->
      union (variables_libres_schma schma) (variables_libres_env reste)

(* Gnralisation d'un type en un schma *)

let gnralisation typ env =
  { quantif = diffrence (variables_libres typ) (variables_libres_env env);
    corps = typ }

(* Appliquer une substitution  un schma *)
(* Pour viter les problmes de capture de variables lies, on renomme
   toutes les variables lies dans le schma en de nouvelles variables *)

let subst_schma phi sigma =
  (* On gnre de nouvelles variables beta1... betaN *)
  let quantif' = List.map nouvelle_variable sigma.quantif in
  (* On construit la substitution [alphai <- betai] o phi *)
  let psi =
    compose (subst (List.map2 (fun alpha beta -> (alpha, Type_var beta))
                              sigma.quantif quantif'))
            phi in
  (* On renvoie le schma V betai. psi(tau) *)
  { quantif = quantif'; corps = psi sigma.corps }

(* Appliquer une substitution  un environnement *)

let subst_env phi env =
  List.map (fun (ident, schma) -> (ident, subst_schma phi schma)) env

(* Type des oprateurs (polymorphes) *)

let type_oprateur op =
  match op with
    "+" | "-" | "*" | "/" ->
      { quantif = [];
        corps = Type_flche(Type_produit(Type_base "int", Type_base "int"),
                            Type_base "int") }
  | "fix" ->
      let alpha = nouvelle_variable() in
      let alpha_f_alpha = Type_flche(Type_var alpha, Type_var alpha) in
      { quantif = [alpha];
        corps = Type_flche(Type_flche(alpha_f_alpha, alpha_f_alpha), alpha_f_alpha) }
  | _ ->
      assert false (* oprateur non support *)

(* L'algorithme W *)

exception Erreur_de_typage

let infrence a =

  (* La variable globale phi. *)

  let phi = ref identit in

  (* L'algorithme rcursif W. *)

  let rec w env a =
    match a with
      Var x ->
	begin try
	  instance_gnrique (List.assoc x env)
	with Not_found ->
	  (* la variable x n'est pas dans le domaine de env *)
	  raise Erreur_de_typage
	end
    | Const n ->
	Type_base "int"
    | Op op ->
	instance_gnrique (type_oprateur op)
    | Fun(x, a1) ->
	let alpha = nouveau() in
	let schma_alpha = {quantif = []; corps = alpha} in
	let tau1 = w ((x, schma_alpha) :: env) a1 in
	Type_flche(alpha, tau1)
    | App(a1, a2) ->
	let tau1 = w env a1 in
	let tau2 = w env a2 in
	let alpha = nouveau() in
	phi := compose (mgu [!phi tau1, Type_flche(!phi tau2, alpha)]) !phi;
	alpha
    | Paire(a1, a2) ->
	let tau1 = w env a1 in
	let tau2 = w env a2 in
	Type_produit (tau1, tau2)
    | Fst a ->
	let tau = w env a in
	let alpha1, alpha2 = nouveau(), nouveau() in
	phi := compose (mgu [!phi tau, Type_produit (alpha1, alpha2)]) !phi;
	alpha1
    | Snd a ->
	let tau = w env a in
	let alpha1, alpha2 = nouveau(), nouveau() in
	phi := compose (mgu [!phi tau, Type_produit (alpha1, alpha2)]) !phi;
	alpha2
    | Let(x, a1, a2) ->
	let tau1 = w env a1 in
	let sigma = gnralisation (!phi tau1) (subst_env !phi env) in
	w ((x, sigma) :: env) a2

  (* On dmarre avec l'environnement vide.  la sortie, on applique la substitution courante au type renvoy par W,
     de faon  obtenir un type qui a un sens. *)
  in
  !phi (w [] a)

(* Pour tester:
infrence (App(Op "+", Paire(Const 1, Const 2)));;
infrence (Let("x", Op "+", App(Var "x", Paire(Const 1, Const 2))));;
infrence (Let("id", Fun("x", Var "x"),
                  Paire(App(Var "id", Const 1), App(Var "id", Var "id"))));;
infrence (Fun("x", Paire(Snd (Var "x"), Fst (Var "x"))));;
infrence (Fun("f", App(Var "f", Var "f")));;
*)
