(* Corrig des exercices de programmation du chapitre 6 *)

(********** Exercice 6.1: l'algorithme d'unification ***************)

(* L'algbre de types *)

type exprtype =
    Type_base of string
  | Type_var of string
  | Type_flche of exprtype * exprtype
  | Type_produit of exprtype * exprtype
  | Type_enreg of exprtype
  | Type_vide                                  (* la range uniformment absente *)
  | Type_champ of string * exprtype * exprtype (* e : tau1; tau2 *)
  | Type_abs
  | Type_pre of exprtype

(* Reprsentation des substitutions par des fonctions des types dans les
   types (comme dans l'exercice 3.3) *)

(* La substitution lmentaire [alpha_i <- tau_i] *)

let subst alpha_tau_list =
  let rec do_subst tau =
    match tau with
      Type_var alpha ->
        begin try
          List.assoc alpha alpha_tau_list
        with Not_found ->
          tau
        end
    | Type_base b -> tau
    | Type_flche(tau1, tau2) -> Type_flche(do_subst tau1, do_subst tau2)
    | Type_produit(tau1, tau2) -> Type_produit(do_subst tau1, do_subst tau2)
    | Type_enreg tau -> Type_enreg(do_subst tau)
    | Type_vide -> Type_vide
    | Type_champ(tiq, tau1, tau2) ->
        Type_champ(tiq, do_subst tau1, do_subst tau2)
    | Type_abs -> Type_abs
    | Type_pre tau -> Type_pre(do_subst tau)
  in
    do_subst

(* La composition de substitutions et la substitution identit *)

let compose subst1 subst2 = fun typ -> subst1(subst2(typ))

let identit = fun typ -> typ

(* Appliquer une substitution  un ensemble d'quations entre types
   (une liste de paires de types) *)

let subst_equations subst c =
  List.map (fun (tau1, tau2) -> (subst tau1, subst tau2)) c

(* Test d'occurrence d'une variable dans un type *)

let rec occurrence alpha tau =
  match tau with
    Type_base b -> false
  | Type_var beta -> alpha = beta
  | Type_flche(tau1, tau2) -> occurrence alpha tau1 || occurrence alpha tau2
  | Type_produit(tau1, tau2) -> occurrence alpha tau1 || occurrence alpha tau2
  | Type_enreg tau -> occurrence alpha tau
  | Type_vide -> false
  | Type_champ(tiq, tau1, tau2) ->
      occurrence alpha tau1 || occurrence alpha tau2
  | Type_abs -> false
  | Type_pre tau -> occurrence alpha tau

(* Gnration de nouvelles variables de types a1, a2, a3, ... *)

let compteur_de_variables = ref 0

let nouvelle_variable _ =
  incr compteur_de_variables;
  "a" ^ string_of_int !compteur_de_variables

let nouveau _ =
  Type_var (nouvelle_variable ())

(* Le calcul de l'unificateur principal *)

exception Non_unifiable

let rec mgu c =
  match c with
    [] -> identit
  | (Type_var alpha, Type_var beta) :: c' when alpha = beta -> mgu c'
  | (Type_var alpha, tau) :: c' when not(occurrence alpha tau) ->
      let s = subst [alpha, tau] in 
      compose (mgu(subst_equations s c')) s
  | (tau, Type_var alpha) :: c' when not(occurrence alpha tau) ->
      let s = subst [alpha, tau] in 
      compose (mgu(subst_equations s c')) s
  | (Type_base b1, Type_base b2) :: c' when b1 = b2 -> mgu c'
  | (Type_flche(tau1, tau2), Type_flche(tau1', tau2')) :: c' ->
      mgu((tau1, tau1') :: (tau2, tau2') :: c')
  | (Type_produit(tau1, tau2), Type_produit(tau1', tau2')) :: c' ->
      mgu((tau1, tau1') :: (tau2, tau2') :: c')
  | (Type_enreg tau, Type_enreg tau') :: c' ->
      mgu((tau, tau') :: c')
  (* Unification entre ranges *)
  | (Type_vide, Type_vide) :: c' -> mgu c'
  | (Type_vide, Type_champ(tiq, tau1, tau2)) :: c' ->
      mgu((tau1, Type_abs) :: (tau2, Type_vide) :: c')
  | (Type_champ(tiq, tau1, tau2), Type_vide) :: c' ->
      mgu((tau1, Type_abs) :: (tau2, Type_vide) :: c')
  | (Type_champ(tiq, tau1, tau2), Type_champ(tiq', tau1', tau2')) :: c' ->
      if tiq = tiq' then
        mgu((tau1, tau1') :: (tau2, tau2') :: c')
      else
        let alpha = nouvelle_variable() in
        mgu((tau2, Type_champ(tiq', tau1', Type_var alpha)) ::
            (tau2', Type_champ(tiq, tau1, Type_var alpha)) :: c')
  (* Unification entre infos de prsence *)
  | (Type_abs, Type_abs) :: c' -> mgu c'
  | (Type_pre tau, Type_pre tau') :: c' ->
      mgu((tau, tau') :: c')
  | _ ->
      raise Non_unifiable

(******************** Exercice 6.2: l'algorithme W ***********************)

(* La syntaxe abstraite des expressions mini-ML avec enregistrements *)

type expression =
    Var of string
  | Const of int
  | Op of string
  | Proj of string                            (* les oprateurs .tiq *)
  | Exten of string                           (* les oprateurs @ tiq *)
  | Fun of string * expression
  | App of expression * expression
  | Paire of expression * expression
  | Fst of expression
  | Snd of expression
  | Let of string * expression * expression
  | Enreg of (string * expression) list       (* {tiq_i = expr_i} *)

(* Les schmas de types (comme au chapitre 1) *)

type schma = { quantif: string list; corps: exprtype }

(* Prendre une instance gnrique d'un schma *)

let instance_gnrique sigma =
  (* On gnre les variables nouvelles beta1... betaN *)
  let vars_inst =
    List.map nouveau sigma.quantif in
  (* On construit la substitution alphai <- betai *)
  let s = subst (List.combine sigma.quantif vars_inst) in
  (* On l'applique au corps du schma *)
  s sigma.corps

(* Calcul des variables libres *)

let rec union l1 l2 =
  match l1 with
    [] -> l2
  | hd :: tl -> union tl (if List.mem hd l2 then l2 else hd :: l2)

let rec variables_libres t =
  match t with
    Type_base b -> []
  | Type_var v -> [v]
  | Type_flche(t1, t2) -> union (variables_libres t1) (variables_libres t2)
  | Type_produit(t1, t2) -> union (variables_libres t1) (variables_libres t2)
  | Type_enreg t -> variables_libres t
  | Type_vide -> []
  | Type_champ(tiq, t1, t2) ->
      union (variables_libres t1) (variables_libres t2)
  | Type_abs -> []
  | Type_pre t -> variables_libres t

let rec diffrence l1 l2 =
  match l1 with
    [] -> []
  | hd :: tl ->
      if List.mem hd l2 then diffrence tl l2 else hd :: diffrence tl l2

let variables_libres_schma s =
  diffrence (variables_libres s.corps) s.quantif

let rec variables_libres_env env =
  match env with
    [] -> []
  | (ident, schma) :: reste ->
      union (variables_libres_schma schma) (variables_libres_env reste)

(* Gnralisation d'un type en un schma *)

let gnralisation typ env =
  { quantif = diffrence (variables_libres typ) (variables_libres_env env);
    corps = typ }

(* Appliquer une substitution  un schma *)
(* Pour viter les problmes de capture de variables lies, on renomme
   toutes les variables lies dans le schma en de nouvelles variables *)

let subst_schma phi sigma =
  (* On gnre de nouvelles variables beta1... betaN *)
  let quantif' = List.map (fun alpha -> nouvelle_variable()) sigma.quantif in
  (* On construit la substitution [alphai <- betai] o phi *)
  let psi =
    compose (subst (List.map2 (fun alpha beta -> (alpha, Type_var beta))
                              sigma.quantif quantif'))
            phi in
  (* On renvoie le schma V betai. psi(tau) *)
  { quantif = quantif'; corps = psi sigma.corps }

(* Appliquer une substitution  un environnement *)

let subst_env phi env =
  List.map (fun (ident, schma) -> (ident, subst_schma phi schma)) env

(* Type des oprateurs (polymorphes) *)

let type_oprateur op =
  match op with
    "+" | "-" | "*" | "/" ->
      { quantif = [];
        corps = Type_flche(Type_produit(Type_base "int", Type_base "int"),
                            Type_base "int") }
  | "fix" ->
      let alpha = nouvelle_variable() in
      { quantif = [alpha];
        corps = Type_flche(Type_flche(Type_var alpha, Type_var alpha),
                            Type_var alpha) }
  | _ ->
      assert false (* oprateur non support *)

(* Type de l'oprateur proj_tiq *)

let type_proj tiq =
  let alpha = nouvelle_variable()
  and beta = nouvelle_variable() in
  { quantif = [alpha; beta];
    corps = Type_flche(Type_enreg(Type_champ(tiq, Type_pre(Type_var alpha),
                                                    Type_var beta)),
                        Type_var alpha) }

(* Type de l'oprateur exten_tiq *)

let type_exten tiq =
  let alpha = nouvelle_variable()
  and beta = nouvelle_variable()
  and gamma = nouvelle_variable() in
  { quantif = [alpha; beta; gamma];
    corps = 
      Type_flche(Type_produit(Type_enreg(Type_champ(tiq, Type_var alpha,
                                                           Type_var beta)),
                               Type_var gamma),
                  Type_enreg(Type_champ(tiq, Type_pre(Type_var gamma),
                                              Type_var beta))) }

(* Tester les doublons; algorithme inefficace *)

exception tiquetteDoublon of string

let rec doublon = function
  | [] ->
      ()
  | tiquette :: suite ->
      if List.mem tiquette suite then
	raise (tiquetteDoublon tiquette);
      doublon suite

(* L'algorithme W *)

exception Erreur_de_typage

let infrence a =

  (* La variable globale phi. *)

  let phi = ref identit in

  (* L'algorithme rcursif W. *)

  let rec w env a =
    match a with
      Var x ->
	begin try
	  instance_gnrique (List.assoc x env)
	with Not_found ->
	  (* la variable x n'est pas dans le domaine de env *)
	  raise Erreur_de_typage
	end
    | Const n ->
	Type_base "int"
    | Op op ->
	instance_gnrique (type_oprateur op)
    | Proj tiq ->
	instance_gnrique (type_proj tiq)
    | Exten tiq ->
	instance_gnrique (type_exten tiq)
    | Fun(x, a1) ->
	let alpha = nouveau() in
	let schma_alpha = {quantif = []; corps = alpha} in
	let tau1 = w ((x, schma_alpha) :: env) a1 in
	Type_flche(alpha, tau1)
    | App(a1, a2) ->
	let tau1 = w env a1 in
	let tau2 = w env a2 in
	let alpha = nouveau() in
	phi := compose (mgu [!phi tau1, Type_flche(!phi tau2, alpha)]) !phi;
	alpha
    | Paire(a1, a2) ->
	let tau1 = w env a1 in
	let tau2 = w env a2 in
	Type_produit (tau1, tau2)
    | Fst a ->
	let tau = w env a in
	let alpha1, alpha2 = nouveau(), nouveau() in
	phi := compose (mgu [!phi tau, Type_produit (alpha1, alpha2)]) !phi;
	alpha1
    | Snd a ->
	let tau = w env a in
	let alpha1, alpha2 = nouveau(), nouveau() in
	phi := compose (mgu [!phi tau, Type_produit (alpha1, alpha2)]) !phi;
	alpha2
    | Let(x, a1, a2) ->
	let tau1 = w env a1 in
	let sigma = gnralisation (!phi tau1) (subst_env !phi env) in
	w ((x, sigma) :: env) a2
    | Enreg tiq_expr_list ->
	doublon (List.map fst tiq_expr_list);
	Type_enreg (
	  List.fold_left (fun range (tiquette, a) ->
	    Type_champ (tiquette, Type_pre (w env a), range)
          ) Type_vide tiq_expr_list
        )

  (* On dmarre avec l'environnement vide.  la sortie, on applique la substitution courante au type renvoy par W,
     de faon  obtenir un type qui a un sens. *)
  in
  !phi (w [] a)

(* Pour tester: compiler

     ocamlc -c records.ml
     ocamlyacc parser.mly
     ocamllex lexer.mll
     ocamlc -c parser.mli
     ocamlc -c parser.ml
     ocamlc -c lexer.ml

   Dans le toplevel ocaml:
     
     #load "records.cmo";;
     #load "parser.cmo";;
     #load "lexer.cmo";;
     open Records;;

     let parse s = (Parser.expression Lexer.token (Lexing.from_string s) : expression);;

     infrence (parse "+(1,2)");;
     infrence (parse "let x = + in x (1,2)");;
     infrence (parse "let id = fun x -> x in (id 1, id id)");;
     infrence (parse "fun x -> (snd x, fst x)");;
     infrence (parse "fun f -> f f");;
     infrence (parse "{ a = 1; a = true }");;
     infrence (parse "{ a = 1; b = (fun x -> x)}");;
     infrence (parse "{ a = 1; b = (fun x -> x)}.a");;
     infrence (parse "{ a = 1; b = (fun x -> x)}.b");;
     infrence (parse "{ a = 1; b = (fun x -> x)}.c");;
     infrence (parse "fun x -> +(x.a, x.b)");;
     infrence (parse "let f = fun x -> +(x.a, x.b) in f { b = 10; a = 1 }");;
     infrence (parse "let f = fun x -> +(x.a, x.b) in f { b = 10 }");;
     infrence (parse "{ a = 1 } @ { b = 2 }");;
     infrence (parse "{ a = 1 } @ { a = (fun x -> x) }");;
     infrence (parse "fun x -> x @ { a = 1 } @ { b = 2 }");;

*)
